#include <stdio.h>
#include <inttypes.h>

#define SOKOL_DEBUG

#define SOKOL_APP_IMPL
#define SOKOL_GFX_IMPL
#define SOKOL_GLUE_IMPL
#define SOKOL_FETCH_IMPL
#define SOKOL_LOG_IMPL
#define SOKOL_METAL
#include <sokol/sokol_log.h>
#include <sokol/sokol_gfx.h>
#include <sokol/sokol_app.h>
#include <sokol/sokol_glue.h>
#include <sokol/sokol_fetch.h>

#define STB_TRUETYPE_IMPLEMENTATION
#include <stb/std_truetype.h>

#define ARENA_IMPLEMENTATION
#include <tsoding/arena.h>

static Arena default_arena = {0};
static Arena temporary_arena = {0};
static Arena *context_arena = &default_arena;

void *context_alloc(size_t size) {
    // assert(context_arena);
    return arena_alloc(context_arena, size);
}

typedef struct {
    float position[2];
    float size[2];
    float border_size[2];
} GpuUiRect;

typedef struct {
    float atlas_position[2];
    float size[2];
    float position[2];
    float y_offset;
} GpuGlyph;

typedef struct {
    float screen_size[2];
} GpuUniformParams;

typedef struct {
    size_t size;
    size_t capacity;
    uint8_t *data;
} U8Array;

U8Array new_u8array(size_t capacity) {
    return (U8Array) {
        .size = 0,
        .capacity = capacity,
        .data = context_alloc(capacity),
    };
}

void push_u8array(U8Array *array, uint8_t *items, size_t num) {
    if (array->size + num <= array->capacity) {
        memcpy(array->data + array->size, items, num);
        array->size += num;
    } else {
        fprintf(stderr, "failed to push to u8 array, size+num > capacity\n");
    }
}

void clear_u8array(U8Array *array) {
    array->size = 0;
}

static struct {
    sg_pass_action pass_action;
    sg_pipeline pip;
    sg_bindings bind;

    sg_image text_atlas_image;
    sg_sampler text_atlas_sampler;

    sg_shader_desc scratch_shader_desc;

    U8Array gpu_ui_rects;
    U8Array gpu_glyphs;
    U8Array glyph_cache;

    bool should_exit;
} state;

#define _String(text) ((String) { .data = text, .len = sizeof(text) })
typedef struct {
    char *data;
    size_t len;
} String;

void queue_text(String text, float position[2]) {
    float x = 0;
    for (size_t i=0; i < text.len; ++i) {
        if (text.data[i] >= 32 && text.data[i] <= 32+95) {
            GpuGlyph glyph = *((GpuGlyph *)(state.glyph_cache.data+((text.data[i] - 32) * sizeof(GpuGlyph))));

            glyph.position[0] = x+position[0];
            glyph.position[1] = position[1];
            x += glyph.size[0]/2;

            push_u8array(&state.gpu_glyphs, &glyph, sizeof(GpuGlyph));
        }
    }
}

void vertex_shader_loaded(const sfetch_response_t *response) {
    if (response->fetched) {
        state.scratch_shader_desc.vs.source = response->data.ptr;
        state.scratch_shader_desc.vs.entry = "vs_main";
        state.scratch_shader_desc.vs = (sg_shader_stage_desc) {
            .source = response->data.ptr,
            .entry = "vs_main",
            .uniform_blocks[0] = {
                .size = sizeof(GpuUniformParams),
                .layout = SG_UNIFORMLAYOUT_NATIVE,
                .uniforms = {
                    [0] = { .name = "screen_size", .type = SG_UNIFORMTYPE_FLOAT2 },
                },
            },
        };
    } else if (response->failed) {
        fprintf(stderr, "failed to load vertex shader\n");
        exit(1);
    }
}

void fragment_shader_loaded(const sfetch_response_t *response) {
    if (response->fetched) {
        state.scratch_shader_desc.fs = (sg_shader_stage_desc){
            .source = response->data.ptr,
            .entry = "fs_main",
            .images[0].used = true,
            .samplers[0].used = true,
            .image_sampler_pairs[0] = { .used = true, .image_slot = 0, .sampler_slot = 0 },
        };
        state.scratch_shader_desc.fs.source = response->data.ptr;
        state.scratch_shader_desc.fs.entry = "fs_main";
    } else if (response->failed) {
        fprintf(stderr, "failed to load vertex shader\n");
        exit(1);
    }
}

void ed_init() {
    uint8_t ttf_buffer[1<<20];
    FILE *ttf_file = fopen("../c_editor/JetBrainsMono-Medium.ttf", "rb");
    if (!ttf_file) {
        exit(1);
    }
    fread(ttf_buffer, 1, 1<<20, ttf_file);
    fclose(ttf_file);

    stbtt_fontinfo font;
    stbtt_InitFont(&font, ttf_buffer, stbtt_GetFontOffsetForIndex(ttf_buffer, 0));

    sg_setup(&(sg_desc) {
        .environment = sglue_environment(),
    });

    float vertices[] = {
        // positions            colors
        -1.0f,  1.0f, 1.0f,     0.0f, 0.0f,
         1.0f,  1.0f, 1.0f,     1.0f, 0.0f,
         1.0f, -1.0f, 1.0f,     1.0f, 1.0f,
        -1.0f, -1.0f, 1.0f,     0.0f, 1.0f,
    };
    state.bind.vertex_buffers[0] = sg_make_buffer(&(sg_buffer_desc) {
        .data = SG_RANGE(vertices)
    });

    const uint16_t indices[] = { 0, 1, 2,  0, 2, 3 };
    state.bind.index_buffer = sg_make_buffer(&(sg_buffer_desc) {
        .type = SG_BUFFERTYPE_INDEXBUFFER,
        .data = SG_RANGE(indices)
    });

    sfetch_setup(&(sfetch_desc_t){ .logger.func = slog_func });

    char vs_source[8000] = { 0 };
    char fs_source[8000] = { 0 };

    sfetch_handle_t vs_handle = sfetch_send(&(sfetch_request_t) {
        .path = "./bin/transpiled_shaders/vertex.metal",
        .callback = vertex_shader_loaded,
        .buffer = { .ptr = vs_source, .size = sizeof(vs_source) },
    });
    sfetch_handle_t fs_handle = sfetch_send(&(sfetch_request_t) {
        .path = "./bin/transpiled_shaders/fragment.metal",
        .callback = fragment_shader_loaded,
        .buffer = { .ptr = fs_source, .size = sizeof(fs_source) },
    });

    // block until files are loaded
    while (sfetch_handle_valid(vs_handle) || sfetch_handle_valid(fs_handle)) {
        sfetch_dowork();
    }

    const int font_bitmap_size = 1024;
    const int rasterized_font_height = 64;
    uint8_t *font_bitmap = context_alloc(font_bitmap_size*font_bitmap_size * sizeof(uint8_t));

    state.gpu_ui_rects = new_u8array(sizeof(GpuUiRect) * 2000);
    state.gpu_glyphs = new_u8array(sizeof(GpuGlyph) * 1024);
    state.glyph_cache = new_u8array(sizeof(GpuGlyph) * 97);

    int ascent, descent, line_gap;
    float scale = stbtt_ScaleForPixelHeight(&font, rasterized_font_height);
    stbtt_GetFontVMetrics(&font, &ascent, &descent, &line_gap);

    for (size_t xx=0; xx < rasterized_font_height/2; ++xx) {
        for (size_t yy=0; yy < rasterized_font_height; ++yy) {
            font_bitmap[xx + yy * font_bitmap_size] = 0;
        }
    }
    // manually add glyph for SPACE
    push_u8array(&state.glyph_cache, &(GpuGlyph){
        .atlas_position = { 0 },
        .size = { rasterized_font_height/4, rasterized_font_height },
        .position = { 0 },
        .y_offset = -rasterized_font_height,
    }, 1 * sizeof(GpuGlyph));

    int x = rasterized_font_height/4;
    int y = 0;
    for (size_t i = 33; i < 33+96; ++i) {
        int width,height, xoff,yoff;
        uint8_t *bitmap = stbtt_GetCodepointBitmap(&font, scale, scale, (int)i, &width, &height, &xoff, &yoff);

        if (x + width >= font_bitmap_size) {
            x = 0;
            y += (int)((float)(ascent - descent + line_gap)*scale);
        }

        int xxx = x;
        for (size_t xx=0; xx < width; ++xx) {
            int yyy = y;
            for (size_t yy=0; yy < height; ++yy) {
                font_bitmap[xxx + yyy * font_bitmap_size] = bitmap[xx + yy * width];

                yyy += 1;
            }

            xxx += 1;
        }

        push_u8array(&state.glyph_cache, &(GpuGlyph){
            .atlas_position = { (float)(x), (float)(y) },
            .size = { (float)width, (float)height },
            .position = { (float)x, (float)y },
            .y_offset = (float)(yoff),
        }, 1 * sizeof(GpuGlyph));

        x += width;
    }

    state.bind.vertex_buffers[1] = sg_make_buffer(&(sg_buffer_desc) {
        .size = state.gpu_glyphs.capacity,
        .usage = SG_USAGE_STREAM,
        .label = "glyph buffer"
    });

    state.text_atlas_sampler = sg_make_sampler(&(sg_sampler_desc) { .mag_filter = SG_FILTER_LINEAR });
    state.text_atlas_image = sg_make_image(&(sg_image_desc) {
        .width = font_bitmap_size,
        .height = font_bitmap_size,
        .pixel_format = SG_PIXELFORMAT_R8,
        .data.subimage[0][0] = { .ptr = font_bitmap, .size = font_bitmap_size*font_bitmap_size * sizeof(uint8_t) },
    });
    state.bind.fs.images[0] = state.text_atlas_image;
    state.bind.fs.samplers[0] = state.text_atlas_sampler;

    state.pass_action = (sg_pass_action) {
        .colors[0] = {
            .load_action = SG_LOADACTION_CLEAR,
            .clear_value = { 0.0f, 0.0f, 0.0f, 1.0f }
        }
    };
    sg_shader shd = sg_make_shader(&state.scratch_shader_desc);
    state.pip = sg_make_pipeline(&(sg_pipeline_desc) {
        .shader = shd,
        .index_type = SG_INDEXTYPE_UINT16,
        .layout = {
            .buffers[1].step_func = SG_VERTEXSTEP_PER_INSTANCE,
            .attrs = {
                [0] = { .format=SG_VERTEXFORMAT_FLOAT3, .buffer_index = 0 },
                [1] = { .format=SG_VERTEXFORMAT_FLOAT2, .buffer_index = 0 },

                [2] = { .format=SG_VERTEXFORMAT_FLOAT2, .buffer_index = 1 },
                [3] = { .format=SG_VERTEXFORMAT_FLOAT2, .buffer_index = 1 },
                [4] = { .format=SG_VERTEXFORMAT_FLOAT2, .buffer_index = 1 },
                [5] = { .format=SG_VERTEXFORMAT_FLOAT, .buffer_index = 1 },
            },
        },
    });

    queue_text(_String("But what even is text! []!@#$%^&*()_=+"), (float[]){ 0, 0 });
    queue_text(_String("v0.1.0"), (float[]){ 32, 128 });
    queue_text(_String("an_editor - what even"), (float[]){ 32, 256 });
}
void ed_frame() {
    if (state.gpu_glyphs.size > 0) {
        sg_update_buffer(state.bind.vertex_buffers[1], &(sg_range) {
            .ptr = state.gpu_glyphs.data,
            .size = state.gpu_glyphs.size
        });
    }

    GpuUniformParams gpu_uniform_params = {
        .screen_size = {
            sapp_width(),
            sapp_height()
        },
    };

    sg_begin_pass(&(sg_pass) { .action = state.pass_action, .swapchain = sglue_swapchain() });
    {
        sg_apply_pipeline(state.pip);
        sg_apply_bindings(&state.bind);
        sg_apply_uniforms(SG_SHADERSTAGE_VS, 0, &(sg_range) { .ptr = &gpu_uniform_params, .size = sizeof(GpuUniformParams) });
        sg_draw(0, 6, state.gpu_glyphs.size/sizeof(GpuGlyph));
    }
    sg_end_pass();
    sg_commit();
}
void ed_cleanup() {
    sfetch_shutdown();
    sg_shutdown();
}
void ed_event(const sapp_event *event) {
    switch (event->type) {
        case SAPP_EVENTTYPE_MOUSE_DOWN:
            if (event->mouse_button == SAPP_MOUSEBUTTON_LEFT) {
                sapp_lock_mouse(true);
            }
            break;

        case SAPP_EVENTTYPE_MOUSE_UP:
            if (event->mouse_button == SAPP_MOUSEBUTTON_LEFT) {
                sapp_lock_mouse(false);
            }
            break;

        default:
            break;
    }
}

sapp_desc sokol_main(int argc, char *argv[]) {
    return (sapp_desc) {
        .width = 640,
        .height = 480,
        .init_cb = ed_init,
        .frame_cb = ed_frame,
        .cleanup_cb = ed_cleanup,
        .event_cb = ed_event,
        .icon.sokol_default = true,
        .logger.func = slog_func,
    };
}
